
/*
	This file is part of miSCellaneous, a program library for SuperCollider 3

	Created: 2015-10-28, version 0.13b
	Copyright (C) 2009-2015 Daniel Mayer 
	Email: 	daniel-mayer@email.de
	URL:	www.daniel-mayer.at

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


+Symbol {

	// for most usage cases this is a CPU-cheap check of bus matching,
	// exception is the very unlikely case of long fx chains, all reading from other Buses too

	miSC_checkPbindFxBusMatch { |sym, firstIsSource = false, secondIsLast = false, firstOtherInArgs, secondOtherInArgs|
		var firstOutputs = SynthDescLib.global[this].outputs;
		var secondInputs = SynthDescLib.global[sym].inputs;
		var secondOutputs = SynthDescLib.global[sym].outputs;
		var firstInputs = SynthDescLib.global[this].inputs;

		var conn, areFirstInputsOk, areFirstOutputsOk, areSecondInputsOk, areSecondOutputsOk,
			firstOutNumberOfChannels, secondInNumberOfChannels;

		areFirstInputsOk = firstIsSource.if {
			firstOtherInArgs.isNil.if {
				firstInputs.size == 0
			}{
				firstInputs.every { |io| firstOtherInArgs.includes(io.startingChannel) }
			}
		}{	// if first node (received symbol) denotes effect, its inputs are checked within other match
			true
		};

		areFirstOutputsOk = ((firstOutputs.size == 1) and: { firstOutputs.first.startingChannel == \out });

		areSecondInputsOk = secondOtherInArgs.isNil.if {
			conn = secondInputs;
			(conn.size == 1) and: { conn.first.startingChannel == \in }
		}{
			(secondInputs.every { |io|
				(io.startingChannel == \in) or: {
					secondOtherInArgs.includes(io.startingChannel) }
			}) and: {
				conn = secondInputs.select { |io| io.startingChannel == \in };
				conn.size == 1
			}
		};

		areSecondOutputsOk = secondIsLast.if {
			(secondOutputs.size == 1) and: { secondOutputs.first.startingChannel == \out }
		}{
			// if second node denotes effect before last effect, its outputs are checked within other match
			true
		};

		((areFirstInputsOk && areFirstOutputsOk && areSecondInputsOk && areSecondOutputsOk) and: {
			(firstOutputs[0].numberOfChannels) <= (conn[0].numberOfChannels)
		}).not.if {
			SimpleInitError("PbindFx bus mismatch, source and fx synths must match " ++
				"in/out conventions: \n\n" ++
				".) \tFor source and fx SynthDefs there must be only one out ugen using 'out' as bus arg. \n" ++
				".) \tSource SynthDefs must not have In ugens, except they are admitted by \\otherInArgs.  \n" ++
				".) \tFx synths must read from in buses with ugen In.ar(in, ...) " ++
				"refering to the bus arg 'in', \n\tthere must not be other In ugens, " ++
				"except they are admitted by \\otherInArgs.\n" ++
				".) \tNumber of out channels of preceeding source / fx synth must not be greater than \n" ++
				"\tthe number of the following fx synth's in channels, this is checked for all connections of an fx graph."
			).throw;
		};
		^conn.first.numberOfChannels
	}
}


+SequenceableCollection {

	// order as receiver
	miSC_getFxBusData { |predecessors, successors, fxOrderPositions, fxEvents, srcInstrument,
		otherInArgs, cleanupDelay, server|
		var cleanupDelaySums = [0], fxBuses, splitBuses,
			splitFactors, fxNum = this.size;

		// 0 (src) can have a split, but last fx in topo order cannot
		splitBuses = {} ! fxNum;
		splitFactors = {} ! fxNum;

		splitFactors[0] = successors[0].asArray.size;
		(splitFactors[0] > 1).if {
			splitBuses[0] = Bus.audio(
				server,
				SynthDescLib.global[srcInstrument].outputs.first.numberOfChannels
			)
		};

		this.do { |fxIndex, i|
			var inSize, inst, prevInst, prevInsts, fxEvent = fxEvents[fxIndex-1], maxPrevCleanupDelaySum = 0;
			inst = fxEvents[fxIndex-1][\fx];

			prevInsts = predecessors[fxIndex].asArray.collect { |j|
				// to get the new cleanupDelaySum, we have to add to the
				// max cleanupDelaySum of the predecessors

				// all cleanupDelaySums are calculated without source cleanupDelay first,
				// which is added afterwards, that way branches with no ins have proper cleanupDelay also (think)
				var k = fxOrderPositions[j];

				(cleanupDelaySums[k] >= maxPrevCleanupDelaySum).if {
					maxPrevCleanupDelaySum = cleanupDelaySums[k]
				};

				(j == 0).if { srcInstrument }{ fxEvents[j - 1][\fx] }
			};

			prevInsts.do { |prevInst|
				inSize = prevInst.miSC_checkPbindFxBusMatch(
					inst,
					i == 0,
					i+1 == fxNum,
					(i == 0).if { otherInArgs }{ fxEvents[this[i-1]-1][\otherInArgs] },
					fxEvents[fxIndex-1][\otherInArgs]
				)
			};

			// is nil if there is no in
			fxBuses = fxBuses.add(inSize.notNil.if { Bus.audio(server, inSize) });

			cleanupDelaySums = cleanupDelaySums.add(
				fxEvent[\cleanupDelay] + maxPrevCleanupDelaySum
			);

			(i < (this.size-1)).if {
				splitFactors[i+1] = successors[fxIndex].asArray.size;
				(splitFactors[i+1] > 1).if {
					splitBuses[i+1] = Bus.audio(
						server,
						SynthDescLib.global[inst].outputs.first.numberOfChannels;
					)
				}
			}
		};
		// adjust cleanupDelaySums to source (see comments above)
		cleanupDelaySums = cleanupDelaySums + cleanupDelay;
		^[fxBuses, splitBuses, splitFactors, cleanupDelaySums]
	}
}


